@@ -252,15 +252,24 @@ open_ldap_connection(const char *ldapServer, int port)
     if (version == -1) {
 	version = LDAP_VERSION2;
     }
-    if (ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &version)
-	!= LDAP_OPT_SUCCESS) {
+    if (ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &version) != LDAP_SUCCESS) {
 	fprintf(stderr, "Could not set LDAP_OPT_PROTOCOL_VERSION %d\n",
 	    version);
 	exit(1);
     }
-    if (use_tls && (version == LDAP_VERSION3) && (ldap_start_tls_s(ld, NULL, NULL) != LDAP_SUCCESS)) {
-	fprintf(stderr, "Could not Activate TLS connection\n");
+    if (use_tls) {
+#ifdef LDAP_OPT_X_TLS
+        if (version == LDAP_VERSION3 && ldap_start_tls_s(ld, NULL, NULL) != LDAP_SUCCESS) {
+	    fprintf(stderr, "Could not Activate TLS connection\n");
+	    exit(1);
+	} else {
+	    fprintf(stderr, "TLS requires LDAP version 3\n");
+	    exit(1);
+	}
+#else
+	fprintf(stderr, "TLS not supported with your LDAP library\n");
 	exit(1);
+#endif
     }
 #endif
     squid_ldap_set_timelimit(ld, timelimit);
@@ -516,7 +525,7 @@ main(int argc, char **argv)
 	rfc1738_unescape(user);
 	rfc1738_unescape(passwd);
 	if (!validUsername(user)) {
-	    printf("ERR\n");
+	    printf("ERR No such user\n");
 	    continue;
 	}
 	tryagain = (ld != NULL);
@@ -530,7 +539,7 @@ main(int argc, char **argv)
 		ld = NULL;
 		goto recover;
 	    }
-	    printf("ERR\n");
+	    printf("ERR %s\n", ldap_err2string(squid_ldap_errno(ld)));
 	} else {
 	    printf("OK\n");
 	}
