@@ -1,6 +1,6 @@
 
 /*
- * $Id: rfc1035.c,v 1.22.2.7 2005/01/17 02:46:29 hno Exp $
+ * $Id: rfc1035.c,v 1.22.2.11 2005/05/11 14:24:28 hno Exp $
  *
  * Low level DNS protocol routines
  * AUTHOR: Duane Wessels
@@ -40,6 +40,7 @@
  */
 
 #include "config.h"
+#include "util.h"
 
 #if HAVE_STDIO_H
 #include <stdio.h>
@@ -82,30 +83,9 @@
 #endif
 
 
-typedef struct _rfc1035_header rfc1035_header;
 
 int rfc1035_errno;
 const char *rfc1035_error_message;
-struct _rfc1035_header {
-    unsigned short id;
-    unsigned int qr:1;
-    unsigned int opcode:4;
-    unsigned int aa:1;
-    unsigned int tc:1;
-    unsigned int rd:1;
-    unsigned int ra:1;
-    unsigned int rcode:4;
-    unsigned short qdcount;
-    unsigned short ancount;
-    unsigned short nscount;
-    unsigned short arcount;
-};
-
-static const char *Alphanum =
-"abcdefghijklmnopqrstuvwxyz"
-"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
-"0123456789";
-
 
 /*
  * rfc1035HeaderPack()
@@ -113,10 +93,10 @@ static const char *Alphanum =
  * Packs a rfc1035_header structure into a buffer.
  * Returns number of octets packed (should always be 12)
  */
-static off_t
-rfc1035HeaderPack(char *buf, size_t sz, rfc1035_header * hdr)
+static int
+rfc1035HeaderPack(char *buf, size_t sz, rfc1035_message * hdr)
 {
-    off_t off = 0;
+    int off = 0;
     unsigned short s;
     unsigned short t;
     assert(sz >= 12);
@@ -158,10 +138,10 @@ rfc1035HeaderPack(char *buf, size_t sz, rfc1035_header * hdr)
  * bytes to follow.  Labels must be smaller than 64 octets.
  * Returns number of octets packed.
  */
-static off_t
+static int
 rfc1035LabelPack(char *buf, size_t sz, const char *label)
 {
-    off_t off = 0;
+    int off = 0;
     size_t len = label ? strlen(label) : 0;
     if (label)
 	assert(!strchr(label, '.'));
@@ -183,18 +163,18 @@ rfc1035LabelPack(char *buf, size_t sz, const char *label)
  * Note message compression is not supported here.
  * Returns number of octets packed.
  */
-static off_t
+static int
 rfc1035NamePack(char *buf, size_t sz, const char *name)
 {
-    off_t off = 0;
-    char *copy = strdup(name);
+    int off = 0;
+    char *copy = xstrdup(name);
     char *t;
     /*
      * NOTE: use of strtok here makes names like foo....com valid.
      */
     for (t = strtok(copy, "."); t; t = strtok(NULL, "."))
 	off += rfc1035LabelPack(buf + off, sz - off, t);
-    free(copy);
+    xfree(copy);
     off += rfc1035LabelPack(buf + off, sz - off, NULL);
     assert(off <= sz);
     return off;
@@ -206,14 +186,14 @@ rfc1035NamePack(char *buf, size_t sz, const char *name)
  * Packs a QUESTION section of a message.
  * Returns number of octets packed.
  */
-static off_t
+static int
 rfc1035QuestionPack(char *buf,
     size_t sz,
     const char *name,
     unsigned short type,
     unsigned short class)
 {
-    off_t off = 0;
+    int off = 0;
     unsigned short s;
     off += rfc1035NamePack(buf + off, sz - off, name);
     s = htons(type);
@@ -229,8 +209,8 @@ rfc1035QuestionPack(char *buf,
 /*
  * rfc1035HeaderUnpack()
  * 
- * Unpacks a RFC1035 message header buffer into a rfc1035_header
- * structure.
+ * Unpacks a RFC1035 message header buffer into the header fields
+ * of the rfc1035_message structure.
  *
  * Updates the buffer offset, which is the same as number of
  * octects unpacked since the header starts at offset 0.
@@ -238,7 +218,7 @@ rfc1035QuestionPack(char *buf,
  * Returns 0 (success) or 1 (error)
  */
 static int
-rfc1035HeaderUnpack(const char *buf, size_t sz, off_t * off, rfc1035_header * h)
+rfc1035HeaderUnpack(const char *buf, size_t sz, int *off, rfc1035_message * h)
 {
     unsigned short s;
     unsigned short t;
@@ -299,9 +279,9 @@ rfc1035HeaderUnpack(const char *buf, size_t sz, off_t * off, rfc1035_header * h)
  * Returns 0 (success) or 1 (error)
  */
 static int
-rfc1035NameUnpack(const char *buf, size_t sz, off_t * off, unsigned short *rdlength, char *name, size_t ns, int rdepth)
+rfc1035NameUnpack(const char *buf, size_t sz, int *off, unsigned short *rdlength, char *name, size_t ns, int rdepth)
 {
-    off_t no = 0;
+    int no = 0;
     unsigned char c;
     size_t len;
     assert(ns > 0);
@@ -311,8 +291,8 @@ rfc1035NameUnpack(const char *buf, size_t sz, off_t * off, unsigned short *rdlen
 	if (c > 191) {
 	    /* blasted compression */
 	    unsigned short s;
-	    off_t ptr;
-	    if (rdepth > 64)		/* infinite pointer loop */
+	    int ptr;
+	    if (rdepth > 64)	/* infinite pointer loop */
 		return 1;
 	    memcpy(&s, buf + (*off), sizeof(s));
 	    s = ntohs(s);
@@ -367,12 +347,12 @@ rfc1035NameUnpack(const char *buf, size_t sz, off_t * off, unsigned short *rdlen
  * Returns 0 (success) or 1 (error)
  */
 static int
-rfc1035RRUnpack(const char *buf, size_t sz, off_t * off, rfc1035_rr * RR)
+rfc1035RRUnpack(const char *buf, size_t sz, int *off, rfc1035_rr * RR)
 {
     unsigned short s;
     unsigned int i;
     unsigned short rdlength;
-    off_t rdata_off;
+    int rdata_off;
     if (rfc1035NameUnpack(buf, sz, off, NULL, RR->name, RFC1035_MAXHOSTNAMESZ, 0)) {
 	RFC1035_UNPACK_DEBUG;
 	memset(RR, '\0', sizeof(*RR));
@@ -411,26 +391,26 @@ rfc1035RRUnpack(const char *buf, size_t sz, off_t * off, rfc1035_rr * RR)
     RR->rdlength = rdlength;
     switch (RR->type) {
     case RFC1035_TYPE_PTR:
-	RR->rdata = malloc(RFC1035_MAXHOSTNAMESZ);
+	RR->rdata = xmalloc(RFC1035_MAXHOSTNAMESZ);
 	rdata_off = *off;
-	RR->rdlength = 0; /* Filled in by rfc1035NameUnpack */
+	RR->rdlength = 0;	/* Filled in by rfc1035NameUnpack */
 	if (rfc1035NameUnpack(buf, sz, &rdata_off, &RR->rdlength, RR->rdata, RFC1035_MAXHOSTNAMESZ, 0))
 	    return 1;
-	if (rdata_off != ((*off) + rdlength)) {
+	if (rdata_off > ((*off) + rdlength)) {
 	    /*
 	     * This probably doesn't happen for valid packets, but
 	     * I want to make sure that NameUnpack doesn't go beyond
 	     * the RDATA area.
 	     */
 	    RFC1035_UNPACK_DEBUG;
-	    free(RR->rdata);
+	    xfree(RR->rdata);
 	    memset(RR, '\0', sizeof(*RR));
 	    return 1;
 	}
 	break;
     case RFC1035_TYPE_A:
     default:
-	RR->rdata = malloc(rdlength);
+	RR->rdata = xmalloc(rdlength);
 	memcpy(RR->rdata, buf + (*off), rdlength);
 	break;
     }
@@ -439,15 +419,6 @@ rfc1035RRUnpack(const char *buf, size_t sz, off_t * off, rfc1035_rr * RR)
     return 0;
 }
 
-static unsigned short
-rfc1035Qid(void)
-{
-    static unsigned short qid = 0x0001;
-    if (++qid == 0xFFFF)
-	qid = 0x0001;
-    return qid;
-}
-
 static void
 rfc1035SetErrno(int n)
 {
@@ -485,7 +456,7 @@ rfc1035SetErrno(int n)
     }
 }
 
-void
+static void
 rfc1035RRDestroy(rfc1035_rr * rr, int n)
 {
     if (rr == NULL)
@@ -493,13 +464,87 @@ rfc1035RRDestroy(rfc1035_rr * rr, int n)
     assert(n > 0);
     while (n--) {
 	if (rr[n].rdata)
-	    free(rr[n].rdata);
+	    xfree(rr[n].rdata);
     }
-    free(rr);
+    xfree(rr);
 }
 
 /*
- * rfc1035AnswersUnpack()
+ * rfc1035QueryUnpack()
+ * 
+ * Unpacks a RFC1035 Query Record into 'query' from a message buffer.
+ *
+ * Updates the new message buffer offset.
+ *
+ * Returns 0 (success) or 1 (error)
+ */
+static int
+rfc1035QueryUnpack(const char *buf, size_t sz, int *off, rfc1035_query * query)
+{
+    unsigned short s;
+    if (rfc1035NameUnpack(buf, sz, off, NULL, query->name, RFC1035_MAXHOSTNAMESZ, 0)) {
+	RFC1035_UNPACK_DEBUG;
+	memset(query, '\0', sizeof(*query));
+	return 1;
+    }
+    if (*off + 4 > sz) {
+	RFC1035_UNPACK_DEBUG;
+	memset(query, '\0', sizeof(*query));
+	return 1;
+    }
+    memcpy(&s, buf + *off, 2);
+    *off += 2;
+    query->qtype = ntohs(s);
+    memcpy(&s, buf + *off, 2);
+    *off += 2;
+    query->qclass = ntohs(s);
+    return 0;
+}
+
+void
+rfc1035MessageDestroy(rfc1035_message * msg)
+{
+    if (!msg)
+	return;
+    if (msg->query)
+	xfree(msg->query);
+    if (msg->answer)
+	rfc1035RRDestroy(msg->answer, msg->ancount);
+    xfree(msg);
+}
+
+/*
+ * rfc1035QueryCompare()
+ * 
+ * Compares two rfc1035_query entries
+ *
+ * Returns 0 (equal) or !=0 (different)
+ */
+int
+rfc1035QueryCompare(const rfc1035_query * a, const rfc1035_query * b)
+{
+    size_t la, lb;
+    if (a->qtype != b->qtype)
+	return 1;
+    if (a->qclass != b->qclass)
+	return 1;
+    la = strlen(a->name);
+    lb = strlen(b->name);
+    if (la != lb) {
+	/* Trim root label(s) */
+	while (la > 0 && a->name[la - 1] == '.')
+	    la--;
+	while (lb > 0 && b->name[lb - 1] == '.')
+	    lb--;
+    }
+    if (la != lb)
+	return 1;
+
+    return strncasecmp(a->name, b->name, la);
+}
+
+/*
+ * rfc1035MessageUnpack()
  *
  * Takes the contents of a DNS reply and fills in an array
  * of resource record structures.  The records array is allocated
@@ -510,60 +555,52 @@ rfc1035RRDestroy(rfc1035_rr * rr, int n)
  */
 
 int
-rfc1035AnswersUnpack(const char *buf,
+rfc1035MessageUnpack(const char *buf,
     size_t sz,
-    rfc1035_rr ** records,
-    unsigned short *id)
+    rfc1035_message ** answer)
 {
-    off_t off = 0;
-    int l;
+    int off = 0;
     int i;
     int nr = 0;
-    rfc1035_header hdr;
+    rfc1035_message *msg;
     rfc1035_rr *recs;
-    memset(&hdr, '\0', sizeof(hdr));
-    if (rfc1035HeaderUnpack(buf + off, sz - off, &off, &hdr)) {
+    rfc1035_query *querys;
+    msg = xcalloc(1, sizeof(*msg));
+    if (rfc1035HeaderUnpack(buf + off, sz - off, &off, msg)) {
 	RFC1035_UNPACK_DEBUG;
 	rfc1035SetErrno(rfc1035_unpack_error);
+	xfree(msg);
 	return -rfc1035_unpack_error;
     }
-    *id = hdr.id;
     rfc1035_errno = 0;
     rfc1035_error_message = NULL;
-    if (hdr.rcode) {
+    i = (int) msg->qdcount;
+    if (i != 1) {
+	/* This can not be an answer to our queries.. */
 	RFC1035_UNPACK_DEBUG;
-	rfc1035SetErrno((int) hdr.rcode);
-	return -rfc1035_errno;
+	rfc1035SetErrno(rfc1035_unpack_error);
+	xfree(msg);
+	return -rfc1035_unpack_error;
     }
-    i = (int) hdr.qdcount;
-    /* skip question */
-    while (i--) {
-	do {
-	    l = (int) (unsigned char) *(buf + off);
-	    off++;
-	    if (l > 191) {	/* compression */
-		off++;
-		break;
-	    } else if (l > RFC1035_MAXLABELSZ) {
-		/* illegal combination of compression bits */
-		RFC1035_UNPACK_DEBUG;
-		rfc1035SetErrno(rfc1035_unpack_error);
-		return -rfc1035_unpack_error;
-	    } else {
-		off += l;
-	    }
-	} while (l > 0);	/* a zero-length label terminates */
-	off += 4;		/* qtype, qclass */
-	if (off > sz) {
+    querys = msg->query = xcalloc((int) msg->qdcount, sizeof(*querys));
+    for (i = 0; i < (int) msg->qdcount; i++) {
+	if (rfc1035QueryUnpack(buf, sz, &off, &querys[i])) {
 	    RFC1035_UNPACK_DEBUG;
 	    rfc1035SetErrno(rfc1035_unpack_error);
+	    rfc1035MessageDestroy(msg);
 	    return -rfc1035_unpack_error;
 	}
     }
-    if (hdr.ancount == 0)
+    *answer = msg;
+    if (msg->rcode) {
+	RFC1035_UNPACK_DEBUG;
+	rfc1035SetErrno((int) msg->rcode);
+	return -rfc1035_errno;
+    }
+    if (msg->ancount == 0)
 	return 0;
-    recs = calloc((int)hdr.ancount, sizeof(*recs));
-    for (i = 0; i < (int)hdr.ancount; i++) {
+    recs = msg->answer = xcalloc((int) msg->ancount, sizeof(*recs));
+    for (i = 0; i < (int) msg->ancount; i++) {
 	if (off >= sz) {	/* corrupt packet */
 	    RFC1035_UNPACK_DEBUG;
 	    break;
@@ -579,11 +616,11 @@ rfc1035AnswersUnpack(const char *buf,
 	 * we expected to unpack some answers (ancount != 0), but
 	 * didn't actually get any.
 	 */
-	free(recs);
+	rfc1035MessageDestroy(msg);
+	*answer = NULL;
 	rfc1035SetErrno(rfc1035_unpack_error);
 	return -rfc1035_unpack_error;
     }
-    *records = recs;
     return nr;
 }
 
@@ -595,21 +632,15 @@ rfc1035AnswersUnpack(const char *buf,
  * probably be at least 512 octets.  The 'szp' initially
  * specifies the size of the buffer, on return it contains
  * the size of the message (i.e. how much to write).
- * Return value is the query ID.
+ * Returns the size of the query
  */
-unsigned short
-rfc1035BuildAQuery(const char *hostname, char *buf, size_t * szp)
+ssize_t
+rfc1035BuildAQuery(const char *hostname, char *buf, size_t sz, unsigned short qid, rfc1035_query * query)
 {
-    static rfc1035_header h;
+    static rfc1035_message h;
     size_t offset = 0;
-    size_t sz = *szp;
     memset(&h, '\0', sizeof(h));
-    /* the first char of hostname must be alphanmeric */
-    if (NULL == strchr(Alphanum, *hostname)) {
-	rfc1035SetErrno(3);
-	return 0;
-    }
-    h.id = rfc1035Qid();
+    h.id = qid;
     h.qr = 0;
     h.rd = 1;
     h.opcode = 0;		/* QUERY */
@@ -620,9 +651,13 @@ rfc1035BuildAQuery(const char *hostname, char *buf, size_t * szp)
 	hostname,
 	RFC1035_TYPE_A,
 	RFC1035_CLASS_IN);
+    if (query) {
+	query->qtype = RFC1035_TYPE_A;
+	query->qclass = RFC1035_CLASS_IN;
+	xstrncpy(query->name, hostname, sizeof(query->name));
+    }
     assert(offset <= sz);
-    *szp = (size_t) offset;
-    return h.id;
+    return offset;
 }
 
 /*
@@ -633,14 +668,13 @@ rfc1035BuildAQuery(const char *hostname, char *buf, size_t * szp)
  * probably be at least 512 octets.  The 'szp' initially
  * specifies the size of the buffer, on return it contains
  * the size of the message (i.e. how much to write).
- * Return value is the query ID.
+ * Returns the size of the query
  */
-unsigned short
-rfc1035BuildPTRQuery(const struct in_addr addr, char *buf, size_t * szp)
+ssize_t
+rfc1035BuildPTRQuery(const struct in_addr addr, char *buf, size_t sz, unsigned short qid, rfc1035_query * query)
 {
-    static rfc1035_header h;
+    static rfc1035_message h;
     size_t offset = 0;
-    size_t sz = *szp;
     static char rev[32];
     unsigned int i;
     memset(&h, '\0', sizeof(h));
@@ -650,7 +684,7 @@ rfc1035BuildPTRQuery(const struct in_addr addr, char *buf, size_t * szp)
 	(i >> 8) & 255,
 	(i >> 16) & 255,
 	(i >> 24) & 255);
-    h.id = rfc1035Qid();
+    h.id = qid;
     h.qr = 0;
     h.rd = 1;
     h.opcode = 0;		/* QUERY */
@@ -661,9 +695,13 @@ rfc1035BuildPTRQuery(const struct in_addr addr, char *buf, size_t * szp)
 	rev,
 	RFC1035_TYPE_PTR,
 	RFC1035_CLASS_IN);
+    if (query) {
+	query->qtype = RFC1035_TYPE_PTR;
+	query->qclass = RFC1035_CLASS_IN;
+	xstrncpy(query->name, rev, sizeof(query->name));
+    }
     assert(offset <= sz);
-    *szp = offset;
-    return h.id;
+    return offset;
 }
 
 /*
@@ -671,13 +709,11 @@ rfc1035BuildPTRQuery(const struct in_addr addr, char *buf, size_t * szp)
  * just need a new ID for it.  Lucky for us ID
  * is the first field in the message buffer.
  */
-unsigned short
-rfc1035RetryQuery(char *buf)
+void
+rfc1035SetQueryID(char *buf, unsigned short qid)
 {
-    unsigned short qid = rfc1035Qid();
     unsigned short s = htons(qid);
     memcpy(buf, &s, sizeof(s));
-    return qid;
 }
 
 #if DRIVER
